// Message.h

/* Author:  Marty Kraimer Date:    26MAR1998 */
 
/*************************************************************************
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory, and the Regents of the University of California, as
* Operator of Los Alamos National Laboratory. MPF is distributed
* subject to a Software License Agreement found in file LICENSE that
* is included with this distribution.
*************************************************************************/
#ifndef MessageH
#define MessageH

extern "C" {
int mrr(void);
int msr(const char *name);
int routerInit();
extern int localMessageRouterStart(int location);
extern int tcpMessageRouterClientStart(
    int location,int port,const char *address,int bufsize,int queueSize);
extern int tcpMessageRouterServerStart(
    int location,int port,const char *address,int bufsize,int queueSize);
extern int tcpDebug;
extern int RMRClientDebug;
extern int RMRServerDebug;
}

#include "mpfType.h"
#include "messageType.h"
class DLList;
class DLNode;

class Message;
class ConnectMessage;
class MessagePvt;
class MessageClient;
class MessageServer;
class MessageRouterClient;
class MessageRouterServer;

typedef Message *(*createMessage)();

class MPF;

class Message
{
public:
    messageType	 getType();
    int32 getClientType();
    void setClientType(int32 type);
    int32 getClientExtra();
    void setClientExtra(int32 extra);
    virtual ~Message() {};
    virtual int32  toBuffer(char **buffer) const;
    virtual int32  fromBuffer(const char **buffer);
    virtual int32  fromBufferSwitch(const char **buffer);
    virtual void print() const;
    static Message *allocate(messageType type);
protected:
    Message(messageType mtype) {type = mtype;};
private:
    messageType	type;
    int32 clientType;
    int32 clientExtra;
    MessagePvt	*pvt;
    friend class MPF;
    friend class MessageRouterClient;
    friend class MessageRouterServer;
    friend class MessageClient;
    friend class MessageServer;
    // disallow default copy constructor, assignment operator
    Message(const Message &);
    Message& operator=(const Message&);
};

inline messageType Message::getType() { return type; }
inline int32 Message::getClientType() { return clientType; }
inline void Message::setClientType(int32 ctype) { clientType = ctype;}
inline int32 Message::getClientExtra() { return clientExtra; }
inline void Message::setClientExtra(int32 extra) { clientExtra = extra;}

struct MessageClientPrivate;
typedef void (*clientCallback)(Message *,void *clientPvt);
class MessageClient
{
public:
    MessageClient(clientCallback,void *clientPvt);
    int	bind(const char *server, int location);
    int	send(Message *message);
    void *getClientPvt();
    //Following for routers not for client
    clientCallback clientCall;
    void bindCallback(ConnectMessage *message);
    void	*routerPvt;
private:
    friend class MessageRouterClient;
    void *clientPvt;
    MessageClientPrivate *pvt;
    // disallow default  copy constructor, assignment operator
    MessageClient(const MessageClient &);
    MessageClient& operator=(const MessageClient&);
    //For now disallow destructor
    ~MessageClient();
};

inline void *MessageClient::getClientPvt() {return(clientPvt);}

struct MessageServerPrivate;
typedef Message *(*allocateMessage)();
class MessageServer
{
public:
    MessageServer(const char* name,int queueSize);
    void waitForMessage();
    Message *receive();
    Message *allocReplyMessage(Message *clientMessage,messageType);
    void  queueMessage(Message *);
    int reply(Message *);
    void setQueueSize(int size);
    const char *getName() const;
    void report() const;
private:
    friend class MessageRouterServer;
    MessageServerPrivate *pvt;
    // disallow default  copy constructor, assignment operator
    MessageServer(const MessageServer &);
    MessageServer& operator=(const MessageServer&);
    //For now disallow destructor
    ~MessageServer();
};

class MPF {

public:
  static void *allocate(DLList **ppFreeList,size_t size);
  static void free(DLList *pFreeList,void *pmessage,size_t size);
  static void switchBytes(char *to,const char *from,int nbytes);

  static int32 toBuffer(char **buffer, const bool value);
  static int32 fromBuffer(const char **buffer,bool *pvalue);
  static int32 fromBufferSwitch(const char **buffer,bool *pvalue);

  static int32 toBuffer(char **buffer, const char8 value);
  static int32 fromBuffer(const char **buffer,char8 *pvalue);
  static int32 fromBufferSwitch(const char **buffer,char8 *pvalue);

  static int32 toBuffer(char **buffer, const int16 value);
  static int32 fromBuffer(const char **buffer,int16 *pvalue);
  static int32 fromBufferSwitch(const char **buffer,int16 *pvalue);
  static int32 toBuffer(char **buffer, const int32 value);
  static int32 fromBuffer(const char **buffer,int32 *pvalue);
  static int32 fromBufferSwitch(const char **buffer,int32 *pvalue);

  static int32 toBuffer(char **buffer, const uint16 value);
  static int32 fromBuffer(const char **buffer,uint16 *pvalue);
  static int32 fromBufferSwitch(const char **buffer,uint16 *pvalue);
  static int32 toBuffer(char **buffer, const uint32 value);
  static int32 fromBuffer(const char **buffer,uint32 *pvalue);
  static int32 fromBufferSwitch(const char **buffer,uint32 *pvalue);

  static int32 toBuffer(char **buffer,const float32 value);
  static int32 fromBuffer(const char **buffer,float32 *pvalue);
  static int32 fromBufferSwitch(const char **buffer,float32 *pvalue);
  static int32 toBuffer(char **buffer,const float64 value);
  static int32 fromBuffer(const char **buffer,float64 *pvalue);
  static int32 fromBufferSwitch(const char **buffer,float64 *pvalue);

  static int32 toBuffer(char **buffer,const char *pstring);
  static int32 fromBuffer(const char **buffer,char *pstring,int32 maxlen);
  static int32 fromBufferSwitch(const char **buffer,
      char *pstring,int32 maxlen);

  static int32 toBuffer(char **buffer,const void *,int32 nbytes);
  static int32 fromBuffer(const char **buffer,void *,int32 nbytes);
};

inline void MPF::switchBytes(char *to,const char *from,int nbytes)
{
   for(int i=0; i<nbytes; i++) to[i] = from[nbytes-i-1];
}

inline int32 MPF::toBuffer(char **buffer, const bool bvalue)
{
    char8 value = (bvalue ? 1 : 0);
    memcpy(*buffer,(char *)&value,sizeof(value));
    *buffer += sizeof(value);
    return(sizeof(value));
}
inline int32 MPF::fromBuffer(const char **buffer,bool *pvalue)
{
    char8 value;
    memcpy((char *)&value,*buffer,sizeof(value));
    *buffer += sizeof(value);
    *pvalue = ((value==1)?true:false);
    return(sizeof(value));
}
inline int32 MPF::fromBufferSwitch(const char **buffer,bool *pvalue)
{//char8 does not need byte swap
    return(MPF::fromBuffer(buffer,pvalue));
}

inline int32 MPF::toBuffer(char **buffer, const char8 value)
{
    memcpy(*buffer,(char *)&value,sizeof(value));
    *buffer += sizeof(value);
    return(sizeof(value));
}
inline int32 MPF::fromBuffer(const char **buffer,char8 *pvalue)
{
    memcpy((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}
inline int32 MPF::fromBufferSwitch(const char **buffer,char8 *pvalue)
{//char8 does not need byte swap
    return(MPF::fromBuffer(buffer,pvalue));
}

inline int32 MPF::toBuffer(char **buffer, const int16 value)
{
    memcpy(*buffer,(char *)&value,sizeof(value));
    *buffer += sizeof(value);
    return(sizeof(value));
}
inline int32 MPF::fromBuffer(const char **buffer,int16 *pvalue)
{
    memcpy((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}
inline int32 MPF::fromBufferSwitch(const char **buffer,int16 *pvalue)
{
    MPF::switchBytes((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}

inline int32 MPF::toBuffer(char **buffer, const int32 value)
{
    memcpy(*buffer,(char *)&value,sizeof(value));
    *buffer += sizeof(value);
    return(sizeof(value));
}
inline int32 MPF::fromBuffer(const char **buffer,int32 *pvalue)
{
    memcpy((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}
inline int32 MPF::fromBufferSwitch(const char **buffer,int32 *pvalue)
{
    MPF::switchBytes((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}

inline int32 MPF::toBuffer(char **buffer, const uint16 value)
{
    memcpy(*buffer,(char *)&value,sizeof(value));
    *buffer += sizeof(value);
    return(sizeof(value));
}
inline int32 MPF::fromBuffer(const char **buffer,uint16 *pvalue)
{
    memcpy((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}
inline int32 MPF::fromBufferSwitch(const char **buffer,uint16 *pvalue)
{
    MPF::switchBytes((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}

inline int32 MPF::toBuffer(char **buffer, const uint32 value)
{
    memcpy(*buffer,(char *)&value,sizeof(value));
    *buffer += sizeof(value);
    return(sizeof(value));
}
inline int32 MPF::fromBuffer(const char **buffer,uint32 *pvalue)
{
    memcpy((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}
inline int32 MPF::fromBufferSwitch(const char **buffer,uint32 *pvalue)
{
    MPF::switchBytes((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}

inline int32 MPF::toBuffer(char **buffer,const float32 value)
{
    memcpy(*buffer,(char *)&value,sizeof(value));
    *buffer += sizeof(value);
    return(sizeof(value));
}
inline int32 MPF::fromBuffer(const char **buffer,float32 *pvalue)
{
    memcpy((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}
inline int32 MPF::fromBufferSwitch(const char **buffer,float32 *pvalue)
{
    MPF::switchBytes((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}

inline int32 MPF::toBuffer(char **buffer,const float64 value)
{
    memcpy(*buffer,(char *)&value,sizeof(value));
    *buffer += sizeof(value);
    return(sizeof(value));
}
inline int32 MPF::fromBuffer(const char **buffer,float64 *pvalue)
{
    memcpy((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}
inline int32 MPF::fromBufferSwitch(const char **buffer,float64 *pvalue)
{
    MPF::switchBytes((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}

#endif /* MessageH */
