// client.cc
/*************************************************************************
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory, and the Regents of the University of California, as
* Operator of Los Alamos National Laboratory. MPF is distributed
* subject to a Software License Agreement found in file LICENSE that
* is included with this distribution.
*************************************************************************/

#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>

#include <epicsEvent.h>
#include <epicsThread.h>
#include <cantProceed.h>

#include "Message.h"
#include "ConnectMessage.h"
#include "Int32Message.h"
#include "DLList.h"
#include "test.h"

int clientInt32Debug = 0;
int clientTimeout = 5;

typedef struct privateData {
    connectStatus connectState;
    int32	valueSent;
    int32	valueReceived;
    epicsEventId sem;
}privateData;

static void myCallback(Message *message,void *clientPvt)
{
    privateData *pprivateData = (privateData *)clientPvt;
    switch(message->getType()) {
    case messageTypeConnect:
	{
	    ConnectMessage *pConnectMessage = (ConnectMessage *)message;

	    printf("clientInt32 callback messageTypeConnect status %s\n",
		connectStatusString[pConnectMessage->status]);
	    pprivateData->connectState = pConnectMessage->status;
	    delete pConnectMessage;
            epicsEventSignal(pprivateData->sem);
	}
	break;
    case messageTypeInt32:
	{
            Int32Message *pInt32Message = (Int32Message *)message;
	
	    pprivateData->valueReceived = pInt32Message->value;	
	    if(clientInt32Debug>1) {
		char buffer[100];

		sprintf(buffer,"client value %d pInt32Message %p\n",
		    pprivateData->valueReceived,(void *)pInt32Message);
		printf("%s",buffer);
	    }
	    if(pInt32Message->extra) {
                epicsEventSignal(pprivateData->sem);
            }
	    delete pInt32Message;
	}
	break;
    default:
	printf("got illegal response message\n");
	break;
    }
}

typedef struct args {
    char *server;
    int slot;
    int32 ntoBuffer;
    double delay;
}args;

static void clientInt32(void *inargs)
{
    args *pargs = (args *)inargs;
    const char *server = pargs->server;
    int slot = pargs->slot;
    int32 ntoBuffer = pargs->ntoBuffer;
    double delay = pargs->delay;

    if(ntoBuffer<1) ntoBuffer = 1;
    privateData	*pprivateData = new privateData;
    pprivateData->connectState = connectNo;
    pprivateData->valueReceived = 0;
    pprivateData->valueSent = 0;
    MessageClient *pMessageClient = new MessageClient(myCallback,pprivateData);
    pprivateData->sem  = epicsEventMustCreate(epicsEventEmpty);
    int status = pMessageClient->bind(server,slot);
    if(status) {
	printf("bind returned %d\n",status);
	return;
    }
    while(true) {
	if(pprivateData->connectState!=connectYes) {
	    epicsEventWaitWithTimeout(pprivateData->sem,1.0);
	    continue;
	}
	int nwait=0;
	while(pprivateData->valueSent!=pprivateData->valueReceived) {
	    if(nwait==0)printf("client send %d receive %d\n",
		pprivateData->valueSent,pprivateData->valueReceived);
	    //wait an extra second for more messages
	    epicsEventWaitWithTimeout(pprivateData->sem,1.0);
	    if(nwait++ == 5) break;
	}
	Int32Message *message=0;
	if(pprivateData->valueSent!=pprivateData->valueReceived) {
	    // never got last message sent. Send it again
	    message = new Int32Message;
	    message->value = pprivateData->valueSent;
	    if(clientInt32Debug>1) printf("resending %d\n",message->value);
	    status = pMessageClient->send(message);
	} else for(int i=0; i<ntoBuffer; i++) {
	    message = new Int32Message;
	    message->value = ++pprivateData->valueSent;
            if(i==ntoBuffer-1) message->extra = 1;
	    if(clientInt32Debug) printf("sending %d\n",message->value);
	    status = pMessageClient->send(message);
	    if(status) break;
	}
	if(status) {
	    printf("send returned %d\n",status);
	    delete message;
	    if(delay==0.0) epicsThreadSleep(.01);
	} else {
	    epicsEventMustWait(pprivateData->sem);
	}
	if(delay>0.0) epicsThreadSleep(delay);
    }
}

extern "C" int initClientInt32(const char *server,int slot,
    int32 ntoBuffer,double delay)
{
    args *pargs = (args *)callocMustSucceed(1,sizeof(args),"initClientInt32");
    epicsThreadId threadId;

    pargs->server = (char *)callocMustSucceed(strlen(server)+1,sizeof(char),"initClientInt32");
    strcpy(pargs->server,server);
    pargs->slot = slot;
    pargs->ntoBuffer = ntoBuffer;
    pargs->delay = delay;
    threadId = epicsThreadCreate("clientInt32",epicsThreadPriorityMedium,
        epicsThreadGetStackSize(epicsThreadStackMedium),
        (EPICSTHREADFUNC)clientInt32,(void *)pargs);
    if(!threadId) printf("initClientInt32 epicsThreadCreate Failure\n");
    return(0);
}
